; This procedure loads the proper TOLNet-compliant format template fro surface ozone measurements
; For questions/comments about this routine, contact Thierry Leblanc at leblanc@tmf.jpl.nasa.gov
; but... I will not answer your questions if you do not know IDL programming

pro TOLNetFormat_Surface,version ,ngenheadformat,ngencommformat,ngencommmeta,nrecformat,nrecmeta,genheadlabel,genhead,gencommlabel,reclabel,datalabel,missval,ioformat

; Inputs:
;    "version" (scalar string) = TOLNet format version # (initial version is v1.0)

; Outputs:
;    "ngenheadformat" (scalar integer) = Number of lines in general header section  dedicated to file format defintion and structure (fixed for a given TOLNet format)
;    "ngencommformat" (scalar integer) = Number of lines in general comment section dedicated to file format defintion and structure (fixed for ALL TOLNet formats)
;    "ngencommmeta"   (scalar integer) = Number of lines in general comment section dedicated to meta data (fixed for a given TOLNet format)
;    "nrecformat"     (scalar integer) = Number of lines in record header dedicated to file format defintion and structure (fixed for ALL TOLNet formats)
;    "nrecmeta"       (scalar integer) = Number of lines in record header dedicated to meta data (fixed for a given TOLNet format)
;    "genheadlabel"   (vector string)  = Content of general header section labels (right hand side of separator " ; ")
;    "genhead"        (vector string)  = Content of general header section meta data (left hand side of separator " ; ")
;    "gencommlabel"   (vector string)  = Content of general comment section labels (right hand side of separator " ; ")
;    "reclabel"       (vector string)  = Content of record header labels (right hand side of separator " ; ")
;    "datalabel"      (vector string)  = Content of data columns header (just above data section)
;    "missval"        (scalar string)  = Content of missing values

; Programming history
;-------------------------------------------------------------------------------------------------------------
; 2012-12-04: Subroutine created
; 2012-12-04: TOLNet profile data format version 1.0 as of 29-Nov-2012: 5 general header metadata lines, 10 profile header metadata lines, 14 data columns
; 2016-04-13: Added RHw among the surface meteorology variables, now 11 columns instead of 10 columns
; 2016-04-13: TOLNet profile data format version 1.1 as of 13-Apr-2016: 5 general header metadata lines, 11 profile header metadata lines, 14 data columns
;---------------------------------------------------------------------------------------------------------------------------------------------

if version eq 'v1.0' then begin
  printf,'***** As of April 13, 2016, this version is obsolete, please use latest version. *****' & stop
endif

if version eq 'v1.1' then begin

; Define general header
genheadlabel=[             'NUMBER OF GENERAL HEADER LINES (AFTER THIS LINE)']    & genhead=['']         & ngenheadformat=1
genheadlabel=[genheadlabel,'TOLNET STANDARDIZED FORMAT VERSION FOR SURFACE DATA'] & genhead=[genhead,''] & ngenheadformat=ngenheadformat+1
genheadlabel=[genheadlabel,'NUMBER OF RECORDS IN THIS FILE']                      & genhead=[genhead,''] & ngenheadformat=ngenheadformat+1
genheadlabel=[genheadlabel,'NUMBER OF DATA COLUMNS FOR ALL RECORDS']              & genhead=[genhead,''] & ngenheadformat=ngenheadformat+1
genheadlabel=[genheadlabel,'COLUMN 1' ] & genhead=[genhead,'StartTime, s, Start time (seconds elapsed since 00:00 UT same day)       '] & columnformat='i5'                  & missval=         -9999  & ndatacol=1
genheadlabel=[genheadlabel,'COLUMN 2' ] & genhead=[genhead,'EndTime, s, End time (seconds elapsed since 00:00 UT same day)           '] & columnformat=[columnformat,'i5']   & missval=[missval,-9999] & ndatacol=ndatacol+1
genheadlabel=[genheadlabel,'COLUMN 3' ] & genhead=[genhead,'MeanTime, s, Weighted averaged time between start and end (UT seconds)   '] & columnformat=[columnformat,'i5']   & missval=[missval,-9999] & ndatacol=ndatacol+1
genheadlabel=[genheadlabel,'COLUMN 4' ] & genhead=[genhead,'O3MR, ppbv, Ozone Mixing Ratio (measured)                                '] & columnformat=[columnformat,'f0.2'] & missval=[missval,-9999] & ndatacol=ndatacol+1
genheadlabel=[genheadlabel,'COLUMN 5' ] & genhead=[genhead,'O3MRUncert, ppbv, Ozone Mixing Ratio Combined Standard Uncertainty       '] & columnformat=[columnformat,'f0.2'] & missval=[missval,-9999] & ndatacol=ndatacol+1
genheadlabel=[genheadlabel,'COLUMN 6' ] & genhead=[genhead,'Precision, %, Measurement Precision                                      '] & columnformat=[columnformat,'f0.2'] & missval=[missval,-9999] & ndatacol=ndatacol+1
genheadlabel=[genheadlabel,'COLUMN 7' ] & genhead=[genhead,'Press, hPa, Surface Air Pressure used to derive Ozone Mixing Ratio       '] & columnformat=[columnformat,'e0.3'] & missval=[missval,-9999] & ndatacol=ndatacol+1
genheadlabel=[genheadlabel,'COLUMN 8']  & genhead=[genhead,'Temp, K, Surface Air Temperature used to derive Ozone Mixing Ratio       '] & columnformat=[columnformat,'f0.2'] & missval=[missval,-9999] & ndatacol=ndatacol+1
genheadlabel=[genheadlabel,'COLUMN 9']  & genhead=[genhead,'RHw, %, Surface Relative Humidity with respect water                     '] & columnformat=[columnformat,'f0.1'] & missval=[missval,-9999] & ndatacol=ndatacol+1
genheadlabel=[genheadlabel,'COLUMN 10'] & genhead=[genhead,'WndSpd, m.s-1, Surface Wind Speed                                        '] & columnformat=[columnformat,'f0.2'] & missval=[missval,-9999] & ndatacol=ndatacol+1
genheadlabel=[genheadlabel,'COLUMN 11'] & genhead=[genhead,'WndDir, deg, Surface Wind Direction (clockwise degrees from North)       '] & columnformat=[columnformat,'f0.1'] & missval=[missval,-9999] & ndatacol=ndatacol+1
genheadlabel=[genheadlabel,'MISSING DATA VALUES'] & genhead=[genhead,strjoin(strtrim(missval,2),', ')] & ngenheadformat=ngenheadformat+1

; Define I/O format
ioformat=strarr(7)
ioformat(0)='(a,t8,a4,a)'
ioformat(1)='(a,t75,a4,a)'
ioformat(2)='(a,t'+strtrim(strlen(genhead(n_elements(genhead)-1))+1,2)+',a4,a)'
ioformat(3)='(a,t60,a4,a)'
ioformat(4)='(a,t64,a4,a)'
ioformat(5)='(a,a4,a)'
ioformat(6)='('+strjoin(columnformat,',", ",')+')'

; Define general comment section (excluding optional comments)
gencommlabel=[          'NUMBER OF GENERAL COMMENTS LINES (AFTER THIS LINE)'] & ngencommformat=1
gencommlabel=[gencommlabel,'INSTRUMENT NAME']                                       & ngencommmeta=1
gencommlabel=[gencommlabel,'PI AND CONTACT INFO']                                   & ngencommmeta=ngencommmeta+1
gencommlabel=[gencommlabel,'SITE NAME']                                             & ngencommmeta=ngencommmeta+1
gencommlabel=[gencommlabel,'SITE LONGITUDE, LATITUDE, ELEVATION (degE, degN, m)']   & ngencommmeta=ngencommmeta+1
gencommlabel=[gencommlabel,'DATA REVISION # (if value >0 then provide text below)'] & ngencommmeta=ngencommmeta+1

; Define record header (excluding optional comments)
reclabel=[         'NUMBER OF HEADER LINES IN THIS RECORD"S HEADER (AFTER THIS LINE)'] & nrecformat=1
reclabel=[reclabel,'NUMBER OF DATA LINES IN THIS RECORD']                              & nrecformat=nrecformat+1
reclabel=[reclabel,'DATA PROCESSING DATE, TIME']                                       & nrecmeta=1
reclabel=[reclabel,'DATA PROCESSING VERSION']                                          & nrecmeta=nrecmeta+1
reclabel=[reclabel,'RESULTS QUALITY (NOMINAL, FAIR, POOR)']                            & nrecmeta=nrecmeta+1
reclabel=[reclabel,'RECORD DATE, TIME (UT) START']                                     & nrecmeta=nrecmeta+1
reclabel=[reclabel,'RECORD DATE, TIME (UT) END']                                       & nrecmeta=nrecmeta+1
reclabel=[reclabel,'SOURCE OF SURFACE PRESSURE, TEMPERATURE, RH, AND WIND']            & nrecmeta=nrecmeta+1
reclabel=[reclabel,'OPERATOR COMMENTS']                                                & nrecmeta=nrecmeta+1
datalabel=strarr(ndatacol) & igcol0=min(where(strpos(genheadlabel,'COLUMN 1') ge 0))
for igcol=igcol0,igcol0+ndatacol-1 do begin
  splitgenhead=strsplit(genhead(igcol),',',/extract,/regex) & datalabel(igcol-igcol0)=splitgenhead(0)
endfor
datalabel=strjoin(datalabel,', ') & nrecformat=nrecformat+1

endif

return
end