; This procedure writes out ozone lidar data at the TOLNet-compliant format
; For questions/comments, contact Thierry Leblanc at thierry.leblanc@jpl.nasa.gov
; but... I will not answer your questions if you do not know IDL programming

pro WriteTOLNet,site_id ,instr_name,pi_name,pi_org,pi_email,site_name,site_lon,site_lat,site_alt,rev_id $                                         ; Input parameters (mandatory) for general header
                        ,process_date,process_time,process_v,qual,date_start,time_start,date_end,time_end,date_mean,time_mean $                   ; Input parameters (mandatory) for profile header
                        ,xsce_name,xsce_date,xsce_time,xsce_lon,xsce_lat,xsce_alt,oper_comm   $                                                   ; Input parameters (mandatory) for profile header (continued)
                        ,z,o3nd,uo3nd,zres,prec,chancode,o3mr,uo3mr,p,up,t,ut,d,ud $                                                              ; Input parameters (mandatory) for data columns
                        ,rev_comm=rev_comm,prof_comm=prof_comm,ndata=ndata,create=create,append=append,file_suffix=file_suffix,dir_name=dir_name  ; Input keywords (optional) for I/O options and comments

; Input parameters' and keywords' dimension and type:
;-------------------------------------------------------------------------------------------------------------
; If only 1 profile is being written:
; integer scalar:  rev_id
; string scalar:   site_id, instr_name, pi_name, pi_org, pi_email, site_name, all dates and times, process_v, qual, xsce_name, all xsce_dates and times, oper_comm, file_suffix
; floating scalar: site_lon, site_lat, site_ alt, xsce_lon, xcsce_lat, xsce_alt
; floating vector of length equal to the number of sampled altitudes (ndata): z, o3nd, uo3nd, zres, prec, chancode, o3mr, uo3mr, p, up, t, ut, d, ud
; string vector of length equal to the number of optional comment lines: rev_comm, prof_comm
;-------------------------------------------------------------------------------------------------------------
; If more than 1 profile is being written:
; The "general header" and "general comments" input variables have their dimension unchanged
; All "profile header" input variables have their dimension augmented by one (to be added in front), which typically are vectors of length "nprof" instead of scalars
; All "data" input variables have their dimension augmented by one (to be added in front), which typically are arrays of dimension "nprof x ndata" instead of vectors of length "ndata"
; Note: Use the keyword syntax ",ndata=[..,..,..]" (a vector of length "nprof") to set a different number of data lines for each profile

; Calling options:
;-------------------------------------------------------------------------------------------------------------
; To create a new data file, write its general header, then write one or several profile(s), use the following syntax:
;
; WriteTOLNet,site_id,/create ,instr_name,pi_name,pi_org,pi_email,site_name,site_lon,site_lat,site_alt,rev_id  $
;                             ,process_date,process_time,process_v,qual,date_start,time_start,date_end,time_end,date_mean,time_mean $
;                             ,xsce_name,xsce_date,xsce_time,xsce_lon,xsce_lat,xsce_alt,oper_comm $
;                             ,z,o3nd,uo3nd,zres,prec,chancode,o3mr,uo3mr,p,up,t,ut,d,ud [,prof_comm=...][,rev_comm=...][,ndata=...][,file_suffix=...]

;-------------------------------------------------------------------------------------------------------------
; To append one or several profile(s) to an existing file, use the following syntax:
;
; WriteTOLNet,site_id,/append ,instr_name,pi_name,pi_org,pi_email,site_name,site_lon,site_lat,site_alt,rev_id  $
;                             ,process_date,process_time,process_v,qual,date_start,time_start,date_end,time_end,date_mean,time_mean $
;                             ,xsce_name,xsce_date,xsce_time,xsce_lon,xsce_lat,xsce_alt,oper_comm $
;                             ,z,o3nd,uo3nd,zres,prec,chancode,o3mr,uo3mr,p,up,t,ut,d,ud [,prof_comm=...][,rev_comm=...][,ndata=...][,file_suffix=...]

;-------------------------------------------------------------------------------------------------------------
; Other optional keywords:
; To add a suffix in the filename, use the keyword syntax "file_suffix=..."
; To include revision comment lines, use the optional kewyord syntax "rev_comm=..." as needed
; To include profile comment lines , use the optional keyword syntax "prof_comm=..." as needed
;-------------------------------------------------------------------------------------------------------------

; Programming history
;-------------------------------------------------------------------------------------------------------------
; 2012-12-03: Subroutine created
; 2013-02-04: Added optional keyword "dir_name" and subsequent programming to specifiy output directory
; 2013-02-04: Modified upper limit for valid O3MR and O3MRUncert (from 1.0e4 to 1.5e4 ppbv) for stratospheric profiles
;-------------------------------------------------------------------------------------------------------------

; Check keywords compatibility
;-------------------------------------------------------------------------------------------------------------
if not keyword_set(create) and not keyword_set(append) then message,'One of the following keywords must be used: "/create", or "/append"'
if keyword_set(append) and keyword_set(create)         then message,'Incompatible keywords: use either "/append" or "/create" but not both'
if keyword_set(file_suffix)                            then file_suffix='_'+file_suffix else file_suffix=''
if keyword_set(dir_name)                               then dirname=dir_name else dirname=''
;-------------------------------------------------------------------------------------------------------------

; 1) Check input parameter dimensions
;-------------------------------------------------------------------------------------------------------------
nz=n_elements(z)
if variance([nz,n_elements(o3nd),n_elements(uo3nd),n_elements(zres),n_elements(prec),n_elements(chancode),n_elements(o3mr),n_elements(uo3mr),n_elements(p),n_elements(up),n_elements(t),n_elements(ut),n_elements(d),n_elements(ud)]) ne 0. $
   then message,'Inconsistent dimensions for input data'

ndim=size(z,/n_dimensions)
if ndim ge 3 then message,'Input data dimension must be 1 or 2 only'
if ndim eq 2 then begin
  nprof=n_elements(z(*,0)) & ndatamax=n_elements(z(0,*))
  if variance([nprof,n_elements(o3nd(*,0)),n_elements(uo3nd(*,0)),n_elements(zres(*,0)),n_elements(prec(*,0)),n_elements(chancode(*,0))]) ne 0. or $
     variance([nprof,n_elements(o3mr(*,0)),n_elements(uo3mr(*,0)),n_elements(p(*,0)),n_elements(up(*,0)),n_elements(t(*,0)),n_elements(ut(*,0)),n_elements(d(*,0)),n_elements(ud(*,0))]) ne 0. $
     then message,'Inconsistent number of profiles for input data'

  if variance([nprof,n_elements(process_date),n_elements(process_time),n_elements(process_v),n_elements(qual) $
                    ,n_elements(date_start),n_elements(time_start),n_elements(date_end),n_elements(time_end),n_elements(date_mean),n_elements(time_mean) $
                    ,n_elements(xsce_name),n_elements(xsce_date),n_elements(xsce_time),n_elements(xsce_lon),n_elements(xsce_lat),n_elements(xsce_alt),n_elements(oper_comm)]) ne 0. $
     then message,'Inconsistent number of profiles for input meta data'
endif
if ndim eq 1 then begin
  nprof=1 & ndatamax=nz
  z=reform(z,1,nz) & o3nd=reform(o3nd,1,nz) & uo3nd=reform(uo3nd,1,nz)
  zres=reform(zres,1,nz) & prec=reform(prec,1,nz) & chancode=reform(chancode,1,nz)
  o3mr=reform(o3mr,1,nz) & uo3mr=reform(uo3mr,1,nz)
  p=reform(p,1,nz) & up=reform(up,1,nz) & t=reform(t,1,nz) & ut=reform(ut,1,nz) & d=reform(d,1,nz) & ud=reform(ud,1,nz)
  if keyword_set(ndata) then ndata=reform(ndata,1,n_elements(ndata))
  if keyword_set(prof_comm) then prof_comm=reform(prof_comm,1,n_elements(prof_comm))
endif
if not keyword_set(ndata)     then ndata=replicate(ndatamax,nprof)
if not keyword_set(prof_comm) then prof_comm=replicate('',nprof)

; 2) Load proper TOLNet format template
formatv='v1.0'
TOLNetFormat,formatv ,ngenheadformat,ngencommformat,ngencommmeta,nrecformat,nrecmeta  ,genheadlabel,genhead,gencommlabel,reclabel,datalabel,missval,ioformat
ngenhead=n_elements(genhead)
ngencomm=n_elements(gencommlabel)
nrechead=n_elements(reclabel)
ndatacol=n_elements(missval)


; 3) Fill up general header and general comments with input meta data ("/create" option only)
;-------------------------------------------------------------------------------------------------------------
if keyword_set(create) then begin

for i=0,ngenhead-1 do begin
  if genheadlabel(i) eq 'NUMBER OF GENERAL HEADER LINES (AFTER THIS LINE)'    then genhead(i)=strtrim(ngenhead-1,2)
  if genheadlabel(i) eq 'TOLNET STANDARDIZED FORMAT VERSION FOR PROFILE DATA' then genhead(i)=strtrim(formatv,2)
  if genheadlabel(i) eq 'NUMBER OF PROFILES IN THIS FILE'                     then genhead(i)=strtrim(nprof,2)
  if genheadlabel(i) eq 'NUMBER OF DATA COLUMNS FOR ALL PROFILES'             then genhead(i)=strtrim(ndatacol,2)
  if genheadlabel(i) eq 'MISSING DATA VALUES'                                 then genhead(i)=strjoin(strtrim(missval,2),', ')
endfor

gencomm=strarr(ngencomm)
for i=0,ngencomm-1 do begin
  if gencommlabel(i) eq 'INSTRUMENT NAME'                                       then gencomm(i)=strtrim(instr_name,2)
  if gencommlabel(i) eq 'PI AND CONTACT INFO'                                   then gencomm(i)=strtrim(pi_name,2)+', '+strtrim(pi_org,2)+', '+strtrim(pi_email,2)
  if gencommlabel(i) eq 'SITE NAME'                                             then gencomm(i)=strtrim(site_name,2)
  if gencommlabel(i) eq 'SITE LONGITUDE, LATITUDE, ELEVATION (degE, degN, m)'   then gencomm(i)=strtrim(site_lon,2)+', '+strtrim(site_lat,2)+', '+strtrim(site_alt,2)
  if gencommlabel(i) eq 'DATA REVISION # (if value >0 then provide text below)' then gencomm(i)='R'+strtrim(rev_id,2)
endfor

; Adding optional general header comments
if keyword_set(rev_comm) then begin
  nrevcomm=n_elements(rev_comm)
  for irevcomm=0,nrevcomm-1 do begin
    if irevcomm eq 0 then gencommlabel=[gencommlabel,'DATA REVISION DETAILS, NEWEST ON TOP'] else gencommlabel=[gencommlabel,'']
    gencomm=[gencomm,rev_comm(irevcomm)]
  endfor
endif else begin
  nrevcomm=0
endelse
ngencommx=n_elements(gencomm)

; Finalizing general header and comments with number of general comments lines
for i=0,ngencommx-1 do if gencommlabel(i) eq 'NUMBER OF GENERAL COMMENTS LINES (AFTER THIS LINE)' then gencomm(i)=strtrim(ngencommx-1,2)

endif

; 4) Fill up record headers with input meta data (all options)
;-------------------------------------------------------------------------------------------------------------
nprofcommmax=0 & nrecheadx=indgen(nprof) & nprofcomm=indgen(nprof)
for iprof=0,nprof-1 do begin
  goodprofcomm=where(prof_comm(iprof,*) ne '',ngoodprofcomm) & nprofcomm(iprof)=ngoodprofcomm & nprofcommmax=nprofcommmax>ngoodprofcomm
endfor
nrecheadmax=nrechead+nprofcommmax+1 & recxlabel=strarr(nprof,nrecheadmax) & rechead=recxlabel

for i=0,nrechead-1 do begin
  recxlabel(*,i)=reclabel(i)
  if reclabel(i) eq 'NUMBER OF DATA LINES IN THIS PROFILE'                  then rechead(*,i)=strtrim(ndata,2)
  if reclabel(i) eq 'DATA PROCESSING DATE, TIME'                            then rechead(*,i)=strtrim(process_date,2)+', '+strtrim(process_time,2)
  if reclabel(i) eq 'DATA PROCESSING VERSION'                               then rechead(*,i)=strtrim(process_v,2)
  if reclabel(i) eq 'RESULTS QUALITY (NOMINAL, FAIR, POOR)'                 then rechead(*,i)=strtrim(qual,2)
  if reclabel(i) eq 'PROFILE DATE, TIME (UT) START'                         then rechead(*,i)=strtrim(date_start,2)+', '+strtrim(time_start,2)
  if reclabel(i) eq 'PROFILE DATE, TIME (UT) END'                           then rechead(*,i)=strtrim(date_end,2)  +', '+strtrim(time_end,2)
  if reclabel(i) eq 'PROFILE DATE, TIME (UT) MEAN'                          then rechead(*,i)=strtrim(date_mean,2) +', '+strtrim(time_mean,2)
  if reclabel(i) eq 'SOURCE OF A PRIORI Press, Temp, AirND USED TO DERIVE OZONE MIXING RATIO' then rechead(*,i)=strtrim(xsce_name,2)
  if reclabel(i) eq 'SOURCE DATE, TIME (UT)'                                then rechead(*,i)=strtrim(xsce_date,2) +', '+strtrim(xsce_time,2)
  if reclabel(i) eq 'SOURCE LONGITUDE, LATITUDE, ELEVATION (degE, degN, m)' then rechead(*,i)=strtrim(xsce_lon,2)  +', '+strtrim(xsce_lat,2)+', '+strtrim(xsce_alt)
  if reclabel(i) eq 'OPERATOR COMMENTS'                                     then rechead(*,i)=strtrim(oper_comm,2)
endfor

; Adding optional profile comments
for iprof=0,nprof-1 do begin & if nprofcomm(iprof) gt 0 then begin
  for iprofcomm=0,nprofcomm(iprof)-1 do begin
    if iprofcomm eq 0 then recxlabel(iprof,nrechead+iprofcomm)=['OTHER COMMENTS SPECIFIC TO THIS PROFILE'] else recxlabel(iprof,nrechead+iprofcomm)=''
    rechead(iprof,nrechead+iprofcomm)=prof_comm(iprof,iprofcomm)
  endfor
endif & endfor

; Finalizing profile header with data column labels and number of record header lines
for iprof=0,nprof-1 do begin
  recxlabel(iprof,nrechead+nprofcomm(iprof))='' & rechead(iprof,nrechead+nprofcomm(iprof))=datalabel
  nrecheadx(iprof)=nrechead+nprofcomm(iprof)+1
endfor
for i=0,nrechead-1 do if reclabel(i) eq 'NUMBER OF HEADER LINES IN THIS PROFILE"S HEADER (AFTER THIS LINE)' then rechead(*,i)=strtrim(nrecheadx-1,2)

; 5) Fill up data columns with input data or missing values (all options)
;-------------------------------------------------------------------------------------------------------------
ndatamax=max(ndata) & dataw=dblarr(nprof,ndatacol,ndatamax)
variablename=strtrim(strsplit(datalabel,',',/extract,/regex),2)
for idatacol=0,ndatacol-1 do begin
  if variablename(idatacol) eq 'ALT'         then begin & bad=where(z     lt -421. or z        gt 120000.,nbad) & if nbad gt 0 then z(bad)=missval(idatacol)        & dataw(*,idatacol,*)=z(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'O3ND'        then begin & bad=where(o3nd     le 0. or o3nd     gt 1.e20  ,nbad) & if nbad gt 0 then o3nd(bad)=missval(idatacol)     & dataw(*,idatacol,*)=o3nd(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'O3NDUncert'  then begin & bad=where(uo3nd    le 0. or uo3nd    gt 1.e20  ,nbad) & if nbad gt 0 then uo3nd(bad)=missval(idatacol)    & dataw(*,idatacol,*)=uo3nd(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'O3NDResol'   then begin & bad=where(zres     le 0. or zres     gt 120000.,nbad) & if nbad gt 0 then zres(bad)=missval(idatacol)     & dataw(*,idatacol,*)=zres(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'Precision'   then begin & bad=where(prec     le 0. or prec     gt 200.   ,nbad) & if nbad gt 0 then prec(bad)=missval(idatacol)     & dataw(*,idatacol,*)=prec(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'ChRange'     then begin & bad=where(chancode le 0. or chancode gt 10.    ,nbad) & if nbad gt 0 then chancode(bad)=missval(idatacol) & dataw(*,idatacol,*)=chancode(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'O3MR'        then begin & bad=where(o3mr     le 0. or o3mr     gt 1.5e4  ,nbad) & if nbad gt 0 then o3mr(bad)=missval(idatacol)     & dataw(*,idatacol,*)=o3mr(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'O3MRUncert'  then begin & bad=where(uo3mr    le 0. or uo3mr    gt 1.5e4  ,nbad) & if nbad gt 0 then uo3mr(bad)=missval(idatacol)    & dataw(*,idatacol,*)=uo3mr(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'Press'       then begin & bad=where(p        le 0. or p        gt 1200.  ,nbad) & if nbad gt 0 then p(bad)=missval(idatacol)        & dataw(*,idatacol,*)=p(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'PressUncert' then begin & bad=where(up       le 0. or up       gt 1200.  ,nbad) & if nbad gt 0 then up(bad)=missval(idatacol)       & dataw(*,idatacol,*)=up(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'Temp'        then begin & bad=where(t        le 0. or t        gt 350.   ,nbad) & if nbad gt 0 then t(bad)=missval(idatacol)        & dataw(*,idatacol,*)=t(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'TempUncert'  then begin & bad=where(ut       le 0. or ut       gt 350.   ,nbad) & if nbad gt 0 then ut(bad)=missval(idatacol)       & dataw(*,idatacol,*)=ut(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'AirND'       then begin & bad=where(d        le 0. or d        gt 1.e28  ,nbad) & if nbad gt 0 then d(bad)=missval(idatacol)        & dataw(*,idatacol,*)=d(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'AirNDUncert' then begin & bad=where(ud       le 0. or ud       gt 1.e28  ,nbad) & if nbad gt 0 then ud(bad)=missval(idatacol)       & dataw(*,idatacol,*)=ud(*,0:ndatamax-1) & endif
endfor

; 6) Identify filename based on profile header information
;-------------------------------------------------------------------------------------------------------------
iheaddate=min(where(strpos(recxlabel,'PROFILE DATE, TIME (UT) START') ge 0))
ayy=strmid(rechead(iheaddate),0,4) & amth=strmid(rechead(iheaddate),5,2) & aday=strmid(rechead(iheaddate),8,2)
filename='TOLNet-O3Lidar_'+site_id+'_'+ayy+amth+aday+'_R'+strtrim(rev_id,2)+file_suffix+'.dat


; 7a) "/create" option only: Open file and write out general header
;-------------------------------------------------------------------------------------------------------------
 if keyword_set(create) then begin

  openw,lun,/get_lun,dirname+filename
  for i=0,ngenheadformat-1-1          do printf,lun,genhead(i),' ; ',genheadlabel(i),format=ioformat(0)
  for i=ngenheadformat-1,ngenhead-1-1 do printf,lun,genhead(i),' ; ',genheadlabel(i),format=ioformat(1)
  for i=ngenhead-1,ngenhead-1         do printf,lun,genhead(i),' ; ',genheadlabel(i),format=ioformat(2)
  for i=0,ngencommx-1                 do printf,lun,gencomm(i),' ; ',gencommlabel(i),format=ioformat(3)

endif


; 7b) "/append" option only: Read content of existing file and write out old content in updated file
;-------------------------------------------------------------------------------------------------------------
if keyword_set(append) then begin

  if keyword_set(rev_comm) then begin
    rev_comm=rev_comm & nrevcomm=n_elements(rev_comm)
  endif else begin
    rev_comm=''  & nrevcomm=0
  endelse

; Read existing content
  oldfile=dirname+filename & nlines = file_lines(oldfile) & oldcontent=strarr(nlines)
  openr,oldlun,/get_lun,oldfile
    readf,oldlun,oldcontent
  close,oldlun & free_lun,oldlun
  newcontent=oldcontent

; Update number of profiles in file
  for ilines=0,nlines-1 do begin
    splitcontent=strsplit(oldcontent(ilines),';',/regex,/extract)
    if n_elements(splitcontent) eq 2 then begin & if strtrim(splitcontent(1),2) eq 'NUMBER OF PROFILES IN THIS FILE' then begin
      newcontent(ilines)=strjoin([string(long(strtrim(splitcontent(0),2))+nprof,format='(i-'+strtrim(strlen(splitcontent(0)),2)+')'),splitcontent(1)],';')
    endif & endif
  endfor

; Update revision information in general header (as needed)
  if keyword_set(rev_comm) then begin
  for ilines=0,nlines-1 do begin
    splitcontent=strsplit(oldcontent(ilines),';',/regex,/extract)
    if n_elements(splitcontent) eq 2 then begin & if strtrim(splitcontent(1),2) eq 'NUMBER OF GENERAL COMMENTS LINES (AFTER THIS LINE)' then begin
      newcontent(ilines)=strjoin([string(long(strtrim(splitcontent(0),2))+nrevcomm,format='(i-'+strtrim(strlen(splitcontent(0)),2)+')'),splitcontent(1)],';')
    endif & endif
    if n_elements(splitcontent) eq 2 then begin & if strtrim(splitcontent(1),2) eq 'DATA REVISION # (if value >0 then provide text below)' then begin
      new_rev_id=fix(strmid(splitcontent(0),1,strlen(splitcontent(0))-1))+1
      newcontent(ilines)=strjoin(['R'+string(new_rev_id,format='(i-'+strtrim(strlen(splitcontent(0))-1,2)+')'),splitcontent(1)],';')
      tempcontent=strarr(nlines+nrevcomm)
      tempcontent(0:ilines)=newcontent(0:ilines)
      tempcontent(ilines+1+nrevcomm:nlines+nrevcomm-1)=newcontent(ilines+1:nlines-1)
      for jlines=ilines+1,ilines+1 do tempcontent(jlines)=rev_comm(jlines-ilines-1)+string('  ; DATA REVISION DETAILS (NEWEST ON TOP)',format='(t'+strtrim(60-strlen(rev_comm(jlines-ilines-1)),2)+',a)')
      for jlines=ilines+2,ilines+1+nrevcomm-1 do tempcontent(jlines)=rev_comm(jlines-ilines-1)+string('  ; ',format='(t'+strtrim(60-strlen(rev_comm(jlines-ilines-1)),2)+',a)')
      newcontent=tempcontent
      rev_id=new_rev_id
    endif & endif
  endfor
  endif

; Update finelname based on revision number (as needed)
  newfile='TOLNet-O3Lidar_'+site_id+'_'+ayy+amth+aday+'_R'+strtrim(rev_id,2)+file_suffix+'.dat

; Write updated content
  openw,lun,/get_lun,dirname+newfile
  printf,lun,newcontent

endif


; 8) Write out record header and data
;-------------------------------------------------------------------------------------------------------------
for iprof=0,nprof-1 do begin
  printf,lun,'#BEGIN PROFILE ;----------------------------------------------------------------------------------------------------------'
  for i=0,nrecheadx(iprof)-1-1                do printf,lun,rechead(iprof,i),' ; ',recxlabel(iprof,i) ,format=ioformat(4)
  for i=nrecheadx(iprof)-1,nrecheadx(iprof)-1 do printf,lun,rechead(iprof,i),' ; ',recxlabel(iprof,i) ,format=ioformat(5)
  printf,lun,reform(dataw(iprof,*,0:ndata(iprof)-1)),format=ioformat(6)
endfor

; 8) Close file
;-------------------------------------------------------------------------------------------------------------
close,lun & free_lun,lun

return
end

