; This procedure writes out ozone lidar data at the TOLNet-compliant format
; For questions/comments, contact Thierry Leblanc at leblanc@tmf.jpl.nasa.gov
; but... I will not answer your questions if you do not know IDL programming

pro WriteTOLNet_Surface,site_id ,instr_name,pi_name,pi_org,pi_email,site_name,site_lon,site_lat,site_alt,rev_id $                                ; Input parameters (mandatory) for general header
                       ,process_date,process_time,process_v,qual,date_start,time_start,date_end,time_end,xsce_name,oper_comm   $                 ; Input parameters (mandatory) for record header
                       ,seconds_start,seconds_end,seconds_mean,o3mx,uo3mx,prec,px,tx,rh,wspd,wdir $                                              ; Input parameters (mandatory) for data columns
                       ,rev_comm=rev_comm,rec_comm=rec_comm,ndata=ndata,create=create,append=append,file_suffix=file_suffix,dir_name=dir_name  ; Input keywords (optional) for I/O options and comments

; Input parameters' and keywords' dimension and type:
;-------------------------------------------------------------------------------------------------------------
; If only 1 record is being written:
; integer scalar:  rev_id
; string scalar:   site_id, instr_name, pi_name, pi_org, pi_email, site_name, all dates and times, process_v, qual, xsce_name, oper_comm, file_suffix
; floating scalar: site_lon, site_lat, site_ alt
; floating vector of length equal to the number of sampled altitudes (ndata): seconds_start, seconds_end, seconds_mean, o3mx, uo3mx, prec, px, tx, rh, wspd, wdir
; string vector of length equal to the number of optional comment lines: rev_comm, rec_comm
;-------------------------------------------------------------------------------------------------------------
; If more than 1 record is being written:
; The "general header" and "general comments" input variables have their dimension unchanged
; All "record header" input variables have their dimension augmented by one (to be added in front), which typically are vectors of length "nrec" instead of scalars
; All "data" input variables have their dimension augmented by one (to be added in front), which typically are arrays of dimension "nrec x ndata" instead of vectors of length "ndata"
; Note: Use the keyword syntax ",ndata=[..,..,..]" (a vector of length "nrec") to set a different number of data lines for each record

; Calling options:
;-------------------------------------------------------------------------------------------------------------
; To create a new data file, write its general header, then write one or several record(s), use the following syntax:
;
; WriteTOLNet_Surface,site_id,/create ,instr_name,pi_name,pi_org,pi_email,site_name,site_lon,site_lat,site_alt,rev_id  $
;                                     ,process_date,process_time,process_v,qual,date_start,time_start,date_end,time_end $
;                                     ,xsce_name,oper_comm $
;                                     ,seconds_start,seconds_end,seconds_mean,o3mx,uo3mx,prec,px,tx,rh,wspd,wdir [,rec_comm=...][,rev_comm=...][,ndata=...][,file_suffix=...]
;-------------------------------------------------------------------------------------------------------------
; To append one or several record(s) to an existing file, use the following syntax:
;
; WriteTOLNet_Surface,site_id,/append ,instr_name,pi_name,pi_org,pi_email,site_name,site_lon,site_lat,site_alt,rev_id  $
;                                     ,process_date,process_time,process_v,qual,date_start,time_start,date_end,time_end,date_mean,time_mean $
;                                     ,xsce_name,xsce_date,xsce_time,xsce_lon,xsce_lat,xsce_alt,oper_comm $
;                                     ,seconds_start,seconds_end,seconds_mean,o3mx,uo3mx,prec,px,tx,rh,wspd,wdir [,rec_comm=...][,rev_comm=...][,ndata=...][,file_suffix=...]

;-------------------------------------------------------------------------------------------------------------
; Other optional keywords:
; To add a suffix in the filename, use the keyword syntax "file_suffix=..."
; To include revision comment lines, use the optional kewyord syntax "rev_comm=..." as needed
; To include record comment lines , use the optional keyword syntax "rec_comm=..." as needed
;-------------------------------------------------------------------------------------------------------------

; Programming history
;-------------------------------------------------------------------------------------------------------------
; 2013-09-01: Subroutine created
; 2016-04-13: Sister routine "TOLNetFormat_Surface.pro" modified to include RHw (now 11 columns)
;-------------------------------------------------------------------------------------------------------------

; Check keywords compatibility
;-------------------------------------------------------------------------------------------------------------
if not keyword_set(create) and not keyword_set(append) then message,'One of the following keywords must be used: "/create", or "/append"'
if keyword_set(append) and keyword_set(create)         then message,'Incompatible keywords: use either "/append" or "/create" but not both'
if keyword_set(file_suffix)                            then file_suffix='_'+file_suffix else file_suffix=''
if keyword_set(dir_name)                               then dirname=dir_name else dirname=''
;-------------------------------------------------------------------------------------------------------------

; 1) Check input parameter dimensions
;-------------------------------------------------------------------------------------------------------------
nsample=n_elements(seconds_start)
if variance([nsample,n_elements(seconds_end),n_elements(seconds_mean),n_elements(o3mx),n_elements(uo3mx),n_elements(prec),n_elements(px),n_elements(tx),n_elements(rh),n_elements(wspd),n_elements(wdir)]) ne 0. $
   then message,'Inconsistent dimensions for input data'

ndim=size(seconds_start,/n_dimensions)
if ndim ge 3 then message,'Input data dimension must be 1 or 2 only'
if ndim eq 2 then begin
  nrec=n_elements(seconds_start(*,0)) & ndatamax=n_elements(seconds_start(0,*))
  if variance([nrec,n_elements(seconds_end(*,0)),n_elements(seconds_mean(*,0)),n_elements(o3mx(*,0)),n_elements(uo3mx(*,0)),n_elements(prec(*,0))]) ne 0. or $
     variance([nrec,n_elements(px(*,0)),n_elements(tx(*,0)),n_elements(rh(*,0)),n_elements(wspd(*,0)),n_elements(wdir(*,0))]) ne 0. $
     then message,'Inconsistent number of records for input data'

  if variance([nrec,n_elements(process_date),n_elements(process_time),n_elements(process_v),n_elements(qual) $
                    ,n_elements(date_start),n_elements(time_start),n_elements(date_end),n_elements(time_end),n_elements(date_mean),n_elements(time_mean) $
                    ,n_elements(xsce_name),n_elements(xsce_date),n_elements(xsce_time),n_elements(xsce_lon),n_elements(xsce_lat),n_elements(xsce_alt),n_elements(oper_comm)]) ne 0. $
     then message,'Inconsistent number of records for input meta data'
endif
if ndim eq 1 then begin
  nrec=1 & ndatamax=nsample
  seconds_start=reform(seconds_start,1,nsample) & seconds_end=reform(seconds_end,1,nsample) & seconds_mean=reform(seconds_mean,1,nsample)
  o3mx=reform(o3mx,1,nsample) & uo3mx=reform(uo3mx,1,nsample) & prec=reform(prec,1,nsample)
  px=reform(px,1,nsample) & tx=reform(tx,1,nsample)
  rh=reform(rh,1,nsample) & wspd=reform(wspd,1,nsample) & wdir=reform(wdir,1,nsample)
  if keyword_set(ndata) then ndata=reform(ndata,1,n_elements(ndata))
  if keyword_set(rec_comm) then rec_comm=reform(rec_comm,1,n_elements(rec_comm))
endif
if not keyword_set(ndata)     then ndata=replicate(ndatamax,nrec)
if not keyword_set(rec_comm) then rec_comm=replicate('',nrec)

; 2) Load proper TOLNet format template
formatv='v1.0'
TOLNetFormat_Surface,formatv ,ngenheadformat,ngencommformat,ngencommmeta,nrecformat,nrecmeta  ,genheadlabel,genhead,gencommlabel,reclabel,datalabel,missval,ioformat
ngenhead=n_elements(genhead)
ngencomm=n_elements(gencommlabel)
nrechead=n_elements(reclabel)
ndatacol=n_elements(missval)


; 3) Fill wspd general header and general comments with input meta data ("/create" option only)
;-------------------------------------------------------------------------------------------------------------
if keyword_set(create) then begin

for i=0,ngenhead-1 do begin
  if genheadlabel(i) eq 'NUMBER OF GENERAL HEADER LINES (AFTER THIS LINE)'    then genhead(i)=strtrim(ngenhead-1,2)
  if genheadlabel(i) eq 'TOLNET STANDARDIZED FORMAT VERSION FOR record DATA' then genhead(i)=strtrim(formatv,2)
  if genheadlabel(i) eq 'NUMBER OF RECORDS IN THIS FILE'                     then genhead(i)=strtrim(nrec,2)
  if genheadlabel(i) eq 'NUMBER OF DATA COLUMNS FOR ALL RECORDS'             then genhead(i)=strtrim(ndatacol,2)
  if genheadlabel(i) eq 'MISSING DATA VALUES'                                 then genhead(i)=strjoin(strtrim(missval,2),', ')
endfor

gencomm=strarr(ngencomm)
for i=0,ngencomm-1 do begin
  if gencommlabel(i) eq 'INSTRUMENT NAME'                                       then gencomm(i)=strtrim(instr_name,2)
  if gencommlabel(i) eq 'PI AND CONTACT INFO'                                   then gencomm(i)=strtrim(pi_name,2)+', '+strtrim(pi_org,2)+', '+strtrim(pi_email,2)
  if gencommlabel(i) eq 'SITE NAME'                                             then gencomm(i)=strtrim(site_name,2)
  if gencommlabel(i) eq 'SITE LONGITUDE, LATITUDE, ELEVATION (degE, degN, m)'   then gencomm(i)=strtrim(site_lon,2)+', '+strtrim(site_lat,2)+', '+strtrim(site_alt,2)
  if gencommlabel(i) eq 'DATA REVISION # (if value >0 then provide text below)' then gencomm(i)='R'+strtrim(rev_id,2)
endfor

; Adding optional general header comments
if keyword_set(rev_comm) then begin
  nrevcomm=n_elements(rev_comm)
  for irevcomm=0,nrevcomm-1 do begin
    if irevcomm eq 0 then gencommlabel=[gencommlabel,'DATA REVISION DETAILS, NEWEST ON TOP'] else gencommlabel=[gencommlabel,'']
    gencomm=[gencomm,rev_comm(irevcomm)]
  endfor
endif else begin
  nrevcomm=0
endelse
ngencommx=n_elements(gencomm)

; Finalizing general header and comments with number of general comments lines
for i=0,ngencommx-1 do if gencommlabel(i) eq 'NUMBER OF GENERAL COMMENTS LINES (AFTER THIS LINE)' then gencomm(i)=strtrim(ngencommx-1,2)

endif

; 4) Fill wspd record headers with input meta data (all options)
;-------------------------------------------------------------------------------------------------------------
nreccommmax=0 & nrecheadx=indgen(nrec) & nreccomm=indgen(nrec)
for irec=0,nrec-1 do begin
  goodreccomm=where(rec_comm(irec,*) ne '',ngoodreccomm) & nreccomm(irec)=ngoodreccomm & nreccommmax=nreccommmax>ngoodreccomm
endfor
nrecheadmax=nrechead+nreccommmax+1 & recxlabel=strarr(nrec,nrecheadmax) & rechead=recxlabel

for i=0,nrechead-1 do begin
  recxlabel(*,i)=reclabel(i)
  if reclabel(i) eq 'NUMBER OF DATA LINES IN THIS RECORD'                   then rechead(*,i)=strtrim(ndata,2)
  if reclabel(i) eq 'DATA PROCESSING DATE, TIME'                            then rechead(*,i)=strtrim(process_date,2)+', '+strtrim(process_time,2)
  if reclabel(i) eq 'DATA PROCESSING VERSION'                               then rechead(*,i)=strtrim(process_v,2)
  if reclabel(i) eq 'RESULTS QUALITY (NOMINAL, FAIR, POOR)'                 then rechead(*,i)=strtrim(qual,2)
  if reclabel(i) eq 'RECORD DATE, TIME (UT) START'                          then rechead(*,i)=strtrim(date_start,2)+', '+strtrim(time_start,2)
  if reclabel(i) eq 'RECORD DATE, TIME (UT) END'                            then rechead(*,i)=strtrim(date_end,2)  +', '+strtrim(time_end,2)
  if reclabel(i) eq 'RECORD DATE, TIME (UT) MEAN'                           then rechead(*,i)=strtrim(date_mean,2) +', '+strtrim(time_mean,2)
  if reclabel(i) eq 'SOURCE OF SURFACE PRESSURE, TEMPERATURE, RH, AND WIND' then rechead(*,i)=strtrim(xsce_name,2)
  if reclabel(i) eq 'OPERATOR COMMENTS'                                     then rechead(*,i)=strtrim(oper_comm,2)
endfor

; Adding optional record comments
for irec=0,nrec-1 do begin & if nreccomm(irec) gt 0 then begin
  for ireccomm=0,nreccomm(irec)-1 do begin
    if ireccomm eq 0 then recxlabel(irec,nrechead+ireccomm)=['OTHER COMMENTS SPECIFIC TO THIS RECORD'] else recxlabel(irec,nrechead+ireccomm)=''
    rechead(irec,nrechead+ireccomm)=rec_comm(irec,ireccomm)
  endfor
endif & endfor

; Finalizing record header with data column labels and number of record header lines
for irec=0,nrec-1 do begin
  recxlabel(irec,nrechead+nreccomm(irec))='' & rechead(irec,nrechead+nreccomm(irec))=datalabel
  nrecheadx(irec)=nrechead+nreccomm(irec)+1
endfor
for i=0,nrechead-1 do if reclabel(i) eq 'NUMBER OF HEADER LINES IN THIS RECORD"S HEADER (AFTER THIS LINE)' then rechead(*,i)=strtrim(nrecheadx-1,2)

; 5) Fill wspd data columns with input data or missing values (all options)
;-------------------------------------------------------------------------------------------------------------
ndatamax=max(ndata) & dataw=dblarr(nrec,ndatacol,ndatamax)
variablename=strtrim(strsplit(datalabel,',',/extract,/regex),2)
for idatacol=0,ndatacol-1 do begin
  if variablename(idatacol) eq 'StartTime'  then begin & bad=where(seconds_start lt -3600. or seconds_start ge 86400.,nbad) & if nbad gt 0 then seconds_start(bad)=missval(idatacol) & dataw(*,idatacol,*)=seconds_start(*,0:ndatamax-1) &endif
  if variablename(idatacol) eq 'EndTime'    then begin & bad=where(seconds_end   lt -3600. or seconds_start ge 86400.,nbad) & if nbad gt 0 then seconds_end(bad)=missval(idatacol)   & dataw(*,idatacol,*)=seconds_end(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'MeanTime'   then begin & bad=where(seconds_mean  lt -3600. or seconds_start ge 86400.,nbad) & if nbad gt 0 then seconds_mean(bad)=missval(idatacol)  & dataw(*,idatacol,*)=seconds_mean(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'O3MR'       then begin & bad=where(o3mx          le 0.     or o3mx          gt 1000. ,nbad) & if nbad gt 0 then o3mx(bad)=missval(idatacol)          & dataw(*,idatacol,*)=o3mx(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'O3MRUncert' then begin & bad=where(uo3mx         le 0.     or uo3mx         gt 1000. ,nbad) & if nbad gt 0 then uo3mx(bad)=missval(idatacol)         & dataw(*,idatacol,*)=uo3mx(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'Precision'  then begin & bad=where(prec          le 0.     or prec          gt 200.  ,nbad) & if nbad gt 0 then prec(bad)=missval(idatacol)          & dataw(*,idatacol,*)=prec(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'Press'      then begin & bad=where(px            le 0.     or px            gt 1.5e4 ,nbad) & if nbad gt 0 then px(bad)=missval(idatacol)            & dataw(*,idatacol,*)=px(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'Temp'       then begin & bad=where(tx            le 140    or tx            gt 325   ,nbad) & if nbad gt 0 then tx(bad)=missval(idatacol)            & dataw(*,idatacol,*)=tx(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'RHw'        then begin & bad=where(rh            lt 0.     or rh            gt 120.  ,nbad) & if nbad gt 0 then rh(bad)=missval(idatacol)            & dataw(*,idatacol,*)=rh(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'WndSpd'     then begin & bad=where(wspd          lt 0.     or wspd          gt 200.  ,nbad) & if nbad gt 0 then wspd(bad)=missval(idatacol)          & dataw(*,idatacol,*)=wspd(*,0:ndatamax-1) & endif
  if variablename(idatacol) eq 'WndDir'     then begin & bad=where(wdir          lt 0.     or wdir          ge 360   ,nbad) & if nbad gt 0 then wdir(bad)=missval(idatacol)          & dataw(*,idatacol,*)=wdir(*,0:ndatamax-1) & endif
endfor

; 6) Identify filename based on record header information
;-------------------------------------------------------------------------------------------------------------
iheaddate=min(where(strpos(recxlabel,'RECORD DATE, TIME (UT) START') ge 0))
ayy=strmid(rechead(iheaddate),0,4) & amth=strmid(rechead(iheaddate),5,2) & aday=strmid(rechead(iheaddate),8,2)
filename='TOLNet-O3SurfacePhotometer_'+site_id+'_'+ayy+amth+aday+'_R'+strtrim(rev_id,2)+file_suffix+'.dat'

; 7a) "/create" option only: Open file and write out general header
;-------------------------------------------------------------------------------------------------------------
 if keyword_set(create) then begin

  openw,lun,/get_lun,dirname+filename
  for i=0,ngenheadformat-1-1          do printf,lun,genhead(i),' ; ',genheadlabel(i),format=ioformat(0)
  for i=ngenheadformat-1,ngenhead-1-1 do printf,lun,genhead(i),' ; ',genheadlabel(i),format=ioformat(1)
  for i=ngenhead-1,ngenhead-1         do printf,lun,genhead(i),' ; ',genheadlabel(i),format=ioformat(2)
  for i=0,ngencommx-1                 do printf,lun,gencomm(i),' ; ',gencommlabel(i),format=ioformat(3)

endif


; 7b) "/append" option only: Read content of existing file and write out old content in updated file
;-------------------------------------------------------------------------------------------------------------
if keyword_set(append) then begin

  if keyword_set(rev_comm) then begin
    rev_comm=rev_comm & nrevcomm=n_elements(rev_comm)
  endif else begin
    rev_comm=''  & nrevcomm=0
  endelse

; Read existing content
  oldfile=dirname+filename & nlines = file_lines(oldfile) & oldcontent=strarr(nlines)
  openr,oldlun,/get_lun,oldfile
    readf,oldlun,oldcontent
  close,oldlun & free_lun,oldlun
  newcontent=oldcontent

; Update number of records in file
  for ilines=0,nlines-1 do begin
    splitcontent=strsplit(oldcontent(ilines),';',/regex,/extract)
    if n_elements(splitcontent) eq 2 then begin & if strtrim(splitcontent(1),2) eq 'NUMBER OF RECORDS IN THIS FILE' then begin
      newcontent(ilines)=strjoin([string(long(strtrim(splitcontent(0),2))+nrec,format='(i-'+strtrim(strlen(splitcontent(0)),2)+')'),splitcontent(1)],';')
    endif & endif
  endfor

; Update revision information in general header (as needed)
  if keyword_set(rev_comm) then begin
  for ilines=0,nlines-1 do begin
    splitcontent=strsplit(oldcontent(ilines),';',/regex,/extract)
    if n_elements(splitcontent) eq 2 then begin & if strtrim(splitcontent(1),2) eq 'NUMBER OF GENERAL COMMENTS LINES (AFTER THIS LINE)' then begin
      newcontent(ilines)=strjoin([string(long(strtrim(splitcontent(0),2))+nrevcomm,format='(i-'+strtrim(strlen(splitcontent(0)),2)+')'),splitcontent(1)],';')
    endif & endif
    if n_elements(splitcontent) eq 2 then begin & if strtrim(splitcontent(1),2) eq 'DATA REVISION # (if value >0 then provide text below)' then begin
      new_rev_id=fix(strmid(splitcontent(0),1,strlen(splitcontent(0))-1))+1
      newcontent(ilines)=strjoin(['R'+string(new_rev_id,format='(i-'+strtrim(strlen(splitcontent(0))-1,2)+')'),splitcontent(1)],';')
      tempcontent=strarr(nlines+nrevcomm)
      tempcontent(0:ilines)=newcontent(0:ilines)
      tempcontent(ilines+1+nrevcomm:nlines+nrevcomm-1)=newcontent(ilines+1:nlines-1)
      for jlines=ilines+1,ilines+1 do tempcontent(jlines)=rev_comm(jlines-ilines-1)+string('  ; DATA REVISION DETAILS (NEWEST ON TOP)',format='(t'+strtrim(60-strlen(rev_comm(jlines-ilines-1)),2)+',a)')
      for jlines=ilines+2,ilines+1+nrevcomm-1 do tempcontent(jlines)=rev_comm(jlines-ilines-1)+string('  ; ',format='(t'+strtrim(60-strlen(rev_comm(jlines-ilines-1)),2)+',a)')
      newcontent=tempcontent
      rev_id=new_rev_id
    endif & endif
  endfor
  endif

; Update finelname based on revision number (as needed)
  newfile='TOLNet-O3SurfacePhotometer_'+site_id+'_'+ayy+amth+aday+'_R'+strtrim(rev_id,2)+file_suffix+'.dat

; Write updated content
  openw,lun,/get_lun,dirname+newfile
  printf,lun,newcontent

endif


; 8) Write out record header and data
;-------------------------------------------------------------------------------------------------------------
for irec=0,nrec-1 do begin
  printf,lun,'#BEGIN RECORD ;----------------------------------------------------------------------------------------------------------'
  for i=0,nrecheadx(irec)-1-1                do printf,lun,rechead(irec,i),' ; ',recxlabel(irec,i) ,format=ioformat(4)
  for i=nrecheadx(irec)-1,nrecheadx(irec)-1 do printf,lun,rechead(irec,i),' ; ',recxlabel(irec,i) ,format=ioformat(5)
  printf,lun,reform(dataw(irec,*,0:ndata(irec)-1)),format=ioformat(6)
endfor

; 9) Close file
;-------------------------------------------------------------------------------------------------------------
close,lun & free_lun,lun

return
end

